from __future__ import annotations

from contextlib import asynccontextmanager
from typing import Any, Callable, TypeVar

from pyspec._connection import ClientConnection
from pyspec._connection.data import DataType

from ._motor import Motor
from ._remote_property_table import (
    EventStream,
    Property,
    PropertyGroup,
    RemotePropertyTable,
)
from ._status import Status

T = TypeVar("T", bound=DataType)


class Client(PropertyGroup):
    """
    Represents a client that connects to a PySpec server.

    This is the main entry point for users of the pyspec.client package. Methods allow access to motor control, status monitoring, variable access, and command execution.

    Args:
        host (str): The hostname or IP address of the server.
        port (int): The port number of the server.
    """

    def __init__(self, host: str, port: int):
        self._connection = ClientConnection(host, port)
        self._remote_property_table = RemotePropertyTable(self._connection)
        super().__init__("", self._remote_property_table)

    async def __aenter__(self):
        await self._connection.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc, tb):
        await self._connection.__aexit__(exc_type, exc, tb)

    def status(self) -> Status:
        """
        The status properties reflect changes in the server state that may affect the server's ability to execute client commands or control hardware.

        Returns:
            Status: The status property group. See the pyspec.client._status.Status class for details.
        """
        return Status(self._remote_property_table)

    def motor(self, motor_name: str) -> Motor:
        """
        The motor properties are used to control the motors.
        The parameters for the commands that are sent from the client and the values in the replies and events that are sent from the server are always transmitted as ASCII strings in the data that follows the packet header.

        Args:
            motor_name (str): The name of the motor to control.
        Returns:
            Motor: The motor property group. See the pyspec.client._motor.Motor class for details.
        """
        return Motor(motor_name, self._connection, self._remote_property_table)

    def var(
        self, var_name: str, coerce: Callable[[Any], T] | None = None
    ) -> Property[T]:
        """
        The var properties allow values of any variables to be transferred between the server and the client.
        Enter only the variable name; the property will be created as: var/{var_name}

        var/var_name

        .. code-block:: none

            on("change")
                Sent to clients who have registered when the variable var_name changes value.
            get
                Returns the value of the var_name in the data, if var_name is an existing variable on the server.
            set
                Sets the value of var_name on the server to the contents of data.


        All data types (numbers, strings, associative arrays and data arrays) are supported.

        For built-in associative arrays (A[], S[] and possibly G[], Q[], Z[], U[] and UB[], depending on geometry), only existing elements can be set.

        Properties can be created for individual elements of associative arrays by using the syntax var_name = "array_name[element_key]"

        Args:
            var_name (str): The name of the variable on the server.
            coerce (Callable[[Any], T], optional): An optional function to coerce the data to a specific type.

        Returns:
            :class:`pyspec.client._remote_property_table.Property``[T]`: The property representing the variable on the server.
        """
        return self._property(f"var/{var_name}", coerce)

    def output(self, filename: str) -> EventStream[str]:
        """
        The output property puts copies of the strings written to files or to the screen in events sent to clients.

        output/filename

        .. code-block:: none

            on("change"):
                Sent when the server sends output to the file or device given by filename, where filename can be the built-in name "tty" or a file or device name. The data will be a string representing the output.

            Once a client has registered for output events from a particular file, the server will keep track of the client's request as the file is opened and closed.
            File names are given relative to the server's current directory and can be relative or absolute path names, just as with the built-in commands that refer to files.

            (The output property was introduced in spec release 5.07.04-1.)

        Args:
            filename (str): The file or device name to monitor output for.

        Returns:
            EventStream[str]: An event stream for output events.
        """
        return self._readonly_property(f"output/{filename}", str)

    def count(self) -> Property[bool]:
        """
        The count property provides a count of the number of commands executed by the server since it was started.

        scaler/.all./count
        .. code-block:: none

            on("change")
                Sent when counting starts (data is True) and when counting stops (data is False).
            get
                Data indicates counting (True) or not counting (False).
            set
                If data is nonzero, the server pushes a "count_em data\\\\n" onto the command queue.
                If data is False, counting is aborted as if a ^C had been typed at the server.


        Returns:
            :class:`pyspec.client._remote_property_table.Property``[bool]`: The property representing the count state.
        """
        return self._property("scaler/.all./count", bool)

    async def call(self, function_name: str, *args: str | float | int) -> DataType:
        """
        Call a remote function on the server.

        If this task raises an exception (e.g. due to a timeout),
        the client will send an abort message to the server to stop the execution of the remote function.

        A timeout can be enforced with the following patterns:

        .. code-block:: python

            # Pattern 1: Using asyncio.wait_for
            try:
                result = await asyncio.wait_for(
                    client.call("function_name", arg1, arg2),
                    timeout=5.0,  # seconds
                )
            except asyncio.TimeoutError:
                print("Remote function call timed out.")


            # Pattern 2 (Python 3.11+): Using asyncio.timeout
            try:
                with asyncio.timeout(5.0):  # seconds
                    result = await client.call("function_name", arg1, arg2)
            except asyncio.TimeoutError:
                print("Remote function call timed out.")

        Args:
            function_name (str): The name of the remote function to call.
            *args (str | float | int): The arguments to pass to the remote function.
        Returns:
            DataType: The result of the remote function call.
        """
        return await self._connection.remote_func(function_name, *args)

    async def exec(self, command: str) -> DataType:
        """
        Execute a command on the server.


        If this task raises an exception (e.g. due to a timeout),
        the client will send an abort message to the server to stop the execution of the remote function.

        A timeout can be enforced with the following patterns:

        .. code-block:: python

            # Pattern 1: Using asyncio.wait_for
            try:
                result = await asyncio.wait_for(
                    client.call("function_name", arg1, arg2),
                    timeout=5.0,  # seconds
                )
            except asyncio.TimeoutError:
                print("Remote function call timed out.")


            # Pattern 2 (Python 3.11+): Using asyncio.timeout
            try:
                with asyncio.timeout(5.0):  # seconds
                    result = await client.call("function_name", arg1, arg2)
            except asyncio.TimeoutError:
                print("Remote function call timed out.")

        Args:
            command (str): The command to execute.
        Returns:
            DataType: The result of the command execution.
        """
        return await self._connection.remote_cmd(command)

    @asynccontextmanager
    async def synchronized_motors(self, *, timeout: float | None = None):
        """
        Context manager to enable synchronized motor operations for the client.

        While this context is active, motor movements will be held. Upon exiting the context, the movements will be initialized simultaneously.

        Example usage:

        .. code-block:: python

            async with client.synchronized_motors():
                # Motor movement will be held in here.
                await motor1.move(position)
                await motor2.move(position)

                # Motors will not start moving yet.
                await asyncio.sleep(1)  # Simulate other operations
                # Motors will start moving simultaneously here.

            # Outside of the context, all motors have completed their movements.

        Args:
            timeout (float, optional): Maximum time to wait for all motors to complete, in seconds. If None, wait indefinitely.
        Yields:
            None
        Raises:
            RuntimeError: If there are pending motor motions from a previous context.
        """
        async with self._connection.synchronized_motors(timeout=timeout):
            yield


__all__ = ["Client"]



from typing import Literal, Optional, overload
from pyspec._connection import ClientConnection
import asyncio

from ._remote_property_table import PropertyGroup, RemotePropertyTable, WritableProperty


class Motor(PropertyGroup):
    """
    The motor properties are used to control the motors.
    The parameters for the commands that are sent from the client and the values in the replies and events that are sent from the server are always transmitted as ASCII strings in the data that follows the packet header.

    Args:
        motor_name (str): The name of the motor.
        client_connection (ClientConnection): The client connection instance.
        remote_property_table (RemotePropertyTable): The remote property table instance.
    """

    def __init__(
        self,
        motor_name: str,
        client_connection: ClientConnection,
        remote_property_table: RemotePropertyTable,
    ):
        super().__init__(f"motor/{motor_name}", remote_property_table)
        self.name = motor_name
        self._client_connection = client_connection
        self._remote_property_table = remote_property_table

        self.position = self._readonly_property("position", float)
        """
        motor/{mne}/position
            on("change"): Sent when the dial position or user offset changes.
            get: Returns the current motor position in user units.
            set: Sets the user offset on the server.
        """
        self.dial_position = self._property("dial_position", float)
        """
        motor/{mne}/dial_position
            on("change"): Sent when the dial position changes.
            get: Returns the current motor position in dial units.
            set: Sets the dial position on the server by pushing a

                .. code-block:: none
                
                    set_dial mne data\\n

                onto the command queue, unless the dial position is already set to that value.
        """
        self.offset = self._property("offset", float)
        """
        motor/{mne}/offset
            on("change"): Sent when the offset changes.
            get: Returns the current user offset in dial units.
            set: Sets the user offset by pushing the

                .. code-block:: none
                
                    set mne value\\n

                command onto the command queue, unless the offset is already at the value.
                The data should contain the offset value in motor units (degrees, mm, etc.).
                The server will calculate `value` for the argument in `set` appropriately.
        """
        self.step_size = self._readonly_property("step_size", float)
        """
        motor/{mne}/step_size
            on("change"): Sent when the steps-per-unit parameter changes.
            get: Returns the current steps-per-unit parameter.
        """
        self.sign = self._readonly_property("sign", int)
        """
        motor/{mne}/sign
            on("change"): Sent when the sign-of-user*dial parameter changes.
            get: Returns the current sign-of-user*dial parameter.
        """
        self.moving = self._readonly_property("move_done", bool)
        """
        motor/{mne}/move_done
            on("change"): Sent when moving starts (True) and stops (False).
            get: True if the motor is busy, otherwise False.

        Note: This does seem a little backwards from the name of the SPEC prop.
        """
        self.high_lim_hit = self._readonly_property("high_lim_hit", bool)
        """
        motor/{mne}/high_lim_hit
            on("change"): Sent when the high-limit switch has been hit.
            get: True if the high-limit switch has been hit.
        """
        self.low_lim_hit = self._readonly_property("low_lim_hit", bool)
        """
        motor/{mne}/low_lim_hit
            on("change"): Sent when the low-limit switch has been hit.
            get: True if the low-limit switch has been hit.
        """
        self.emergency_stop = self._readonly_property("emergency_stop", bool)
        """
        motor/{mne}/emergency_stop
            on("change"): Sent when a motor controller indicates a hardware emergency stop.
            get: True if an emergency-stop switch or condition has been activated.
        """
        self.motor_fault = self._readonly_property("motor_fault", bool)
        """
        motor/{mne}/motor_fault
            on("change"): Sent when a motor controller indicates a hardware motor fault.
            get: True if a motor-fault condition has been activated.
        """
        self.high_limit = self._property("high_limit", float)
        """
        motor/{mne}/high_limit
            on("change"): Sent when the value of the high limit position changes.
            get: Returns the high limit in dial units.
            set: Sets the high limit by pushing

                .. code-block:: none
                
                    set_lm  mne data user(mne,get_lim(mne,-1))\\n

                onto the server command queue. (The last argument adds the current low limit to the set_lm command line.)

        """
        self.low_limit = self._property("low_limit", float)
        """
        motor/{mne}/low_limit
            on("change"): Sent when the value of the low limit position changes.
            get: Returns the low limit in dial units.
            set: Sets the low limit by pushing

                .. code-block:: none
                
                    set_lm mne data user(mne,get_lim(mne,+1))\\n

            onto the server command queue. (The last argument adds the current high limit to the set_lm command line.)
        """
        self._limits: WritableProperty[str] = self._writeonly_property("limits")
        """
        motor/{mne}/limits
            set: Sets both motor limits by pushing

                .. code-block:: none
                
                    set_lm mne data\\n

                onto the server command queue,
                where data should contain the low and high motor limit values in a string.
        """
        self._search: WritableProperty[str] = self._writeonly_property("search")
        """
        motor/{mne}/search
            set: The server starts a home or limit search by pushing a

                .. code-block:: none
                
                    chg_dial(mne, how)\\n

                or a
                
                .. code-block:: none
                
                    chg_dial(mne, how, home_pos)\\n

                onto the command queue, depending on whether the data contains one or two arguments.
                The `how` argument is one of the strings recognized by chg_dial(),
                namely \"home\", \"home+\", \"home-\", \"lim+\" or \"lim-\".
                The optional home_pos is the home position in dial units.
        """
        self.unusable = self._readonly_property("unusable", bool)
        """
        motor/{mne}/unusable
            on("change"): Sent when a "disable" option to motor_par() has changed the enabled/disabled state of a motor on the server.
            get: True if the motor is unusable.
        """

        # Sync check is somewhat complicated.
        # This is effectively a request from the server to resolve a synchronization issue.
        # The server will send an event with the current and expected positions,
        # and then wait for a response from either the server keyboard or SOME client.
        # This is not implemented yet, and probably wont be. The use case here is very niche.

        # self.sync_check = self._property("sync_check", str)

        self._start_one: WritableProperty[float] = self._writeonly_property("start_one")
        """
        motor/mne/start_one
            set: If preceded by a prestart_all, adds a

                    
                .. code-block:: none
                
                    A[mne]=data;

                to the buffer that will be pushed onto the server command queue. Otherwise, pushes

                .. code-block:: none
                
                    {get_angles;A[mne]=data;move_em;}\\n

                onto the command queue in order to start the single motor moving.
        """

    async def move(self, position: float):
        """
        Move the motor to the specified position.

        If motor synchronization is enabled, the move will be queued until synchronization is executed.

        Args:
            position (float): The target position to move the motor to.
        """

        if self._client_connection._synchronizing_motors:
            raise RuntimeError(
                "Cannot start move when synchronizing motors. Use enqueue_move instead."
            )

        # Start the tracking before we send the move to avoid race conditions.
        async with self.moving.subscribed(), self.moving.wait_for(False):
            await self._start_one.set(position)

    async def start_move(self, position: float):
        """
        Starts a move to the specified position.
        This will wait for the motor to start moving, and then return a task that can be awaited to wait for the move to complete.

        This is a convenience method that combines move() with waiting for the motor to start moving, which can be useful
        if you want to ensure that the motor is moving before moving on to other tasks.

        Args:
            position (float): The target position to move the motor to.
        """
        async with self.moving.wait_for(True):
            return asyncio.create_task(self.move(position))

    def prepare_move(self, position: float):
        """
        Prepare a move to the specified position without starting it.

        This is used for motor synchronization, where multiple moves are queued and then executed together when synchronization is executed.

        Args:
            position (float): The target position to move the motor to.
        """
        if not self._client_connection._synchronizing_motors:
            raise RuntimeError("Cannot prepare move when not synchronizing motors")

        self._client_connection._pending_motions[self.name] = position

    @overload
    async def search(self, how: Literal["home", "home+", "home-"], home_pos: float): ...

    @overload
    async def search(self, how: Literal["lim+", "lim-"], home_pos: None = None): ...
    async def search(
        self,
        how: Literal["home", "home+", "home-", "lim+", "lim-"],
        home_pos: Optional[float] = None,
    ):
        """
        Start a home or limit search.

        Pushes a chg_dial command like:

        .. code-block:: none

            chg_dial(mne, how)\n
                or
            chg_dial(mne, how, home_pos)\n

        Args:
            how (Literal["home", "home+", "home-", "lim+", "lim-"]): The type of search to perform. Must be one of "home", "home+", "home-", "lim+" or "lim-".
            home_pos (Optional[float]): The home position in dial units (required for home searches).
        """

        if how.startswith("home"):
            if home_pos is None:
                raise ValueError("home_pos must be provided for home searches")
            return await self._search.set(f"{how} {home_pos}")

        return await self._search.set(how)

    async def set_limits(self, low_limit: float, high_limit: float):
        """
        Set both motor limits by pushing:

        .. code-block:: none

            set_lm mne data\n

        onto the server command queue, where data should contain the low and high motor limit values in a string.

        Args:
            low_limit (float): The low limit value.
            high_limit (float): The high limit value.
        """
        await self._limits.set(f"{low_limit} {high_limit}")



from .client_connection import ClientConnection
from .server_connection import ServerConnection
from .associative_array import AssociativeArrayElement, AssociativeArray

__all__ = [
    "ClientConnection",
    "ServerConnection",
    "AssociativeArray",
    "AssociativeArrayElement",
]



import asyncio
import re
import threading
import weakref
from contextlib import asynccontextmanager
from typing import Any, AsyncGenerator, Callable, Dict, Literal, Optional, overload

import numpy as np
from pyee.asyncio import AsyncIOEventEmitter

from pyspec._connection.associative_array import AssociativeArray

from .connection import Connection
from .data import DataType, ErrorStr
from .protocol import Command, Header

LAST_SEQUENCE_NUMBER = 0


class RemoteException(Exception):
    """
    Exception raised when an error occurs on the remote server.
    """


def get_next_sequence_number() -> int:
    """
    Loops through a uint32 sequence number for messages.

    0 is reserved for messages that do not expect a reply.
    1-4294967295 are valid sequence numbers.

    Returns:
        int: The next sequence number.
    """
    global LAST_SEQUENCE_NUMBER
    LAST_SEQUENCE_NUMBER = (LAST_SEQUENCE_NUMBER + 1) % np.iinfo(np.uint32).max
    if LAST_SEQUENCE_NUMBER == 0:
        return get_next_sequence_number()
    return LAST_SEQUENCE_NUMBER


class IndexedSingleton:
    """
    Metaclass for creating indexed singleton classes.

    Each unique combination of __init__ arguments will result in a single instance of the class.
    A weak reference to each instance is stored to allow for garbage collection when no longer in use.
    """

    _instances = {}
    _lock = threading.Lock()

    def __new__(
        cls,
        *args,
    ):
        key = args
        ref = cls._instances.get(key)
        instance = ref() if ref is not None else None
        if instance is None:
            with cls._lock:
                instance = super().__new__(cls)
            cls._instances[key] = weakref.ref(instance)
        return instance


class ClientConnectionEventEmitter(AsyncIOEventEmitter):
    """
    Defines the typed events emitted by the ClientConnection.
    """

    @overload
    def emit(
        self, event: Literal["message"], msg: "ClientConnection.Message"
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["message"],
        func: Callable[["ClientConnection.Message"], Any],
    ) -> Callable[["ClientConnection.Message"], Any]:
        """
        Register an event listener for the 'message' event.

        A 'message' event is emitted whenever a new message is received from the server.
        """

    @overload
    def emit(
        self, event: Literal["property-change"], property_name: str, value: DataType
    ) -> bool: ...
    @overload
    def on(
        self,
        event: Literal["property-change"],
        func: Callable[[str, DataType], Any],
    ) -> Callable[[str, DataType], Any]:
        """
        Register an event listener for the 'property-change' event.

        A 'property-change' event is emitted whenever a property value changes on the server.
        'property-change' events are only emitted for properties that are subscribed to.
        """

    # This is for reply-{sequence_number} events
    @overload
    def emit(self, event: str, data: DataType) -> bool: ...
    @overload
    def on(
        self, event: str, func: Optional[Callable[[DataType], Any]] = None
    ) -> Optional[Callable[[DataType], Any]]:
        """
        Register an event listener for 'reply-{sequence_number}' events.

        A 'reply-{sequence_number}' event is emitted when a reply is received from the server
        corresponding to a command previously sent by the client with that sequence number.
        """

    def emit(self, event: str, *args: Any) -> bool:  # type: ignore[override]
        return super().emit(event, *args)

    def on(self, event: str, func: Optional[Callable[..., Any]] = None):  # type: ignore[override]
        if func is None:
            return super().on(event)
        return super().on(event, func)


class ClientConnection(
    Connection,
    ClientConnectionEventEmitter,
    IndexedSingleton,
):
    """
    Represents a connection to a remote Spec server.

    Provides methods to interact with the server, including reading and writing properties,
    executing commands and functions, and handling events.
    """

    def __init__(self, host: str, port: int) -> None:
        Connection.__init__(self, host, port)
        self.on("message", self._dispatch_typed_message_events)
        self._synchronizing_motors = False
        self._pending_motions: Dict[str, float] = {}

    async def _dispatch_typed_message_events(self, msg: Connection.Message) -> None:
        """
        Given a received message, emit the appropriate typed event based on the message command.

        Args:
            msg (Connection.Message): The received message.
        """
        if msg.header.command == Command.EVENT:
            self.emit("property-change", msg.header.name, msg.data)
        elif (
            msg.header.command == Command.REPLY
            or msg.header.command == Command.HELLO_REPLY
        ):
            self.emit(f"reply-{msg.header.sequence_number}", msg.data)
        else:
            self.logger.error(
                "Received message with unrecognized command: %s", msg.header.command
            )

    async def __aenter__(self) -> "ClientConnection":
        self._reader, self._writer = await asyncio.open_connection(self.host, self.port)
        await super().__aenter__()

        self.logger.info("Connected")

        await self.hello()

        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        await super().__aexit__(exc_type, exc, tb)

        if self.is_connected:
            await self._send(Header(Command.CLOSE))
        if self._writer:
            self._writer.close()
            await self._writer.wait_closed()

    async def _send_with_reply(self, header: Header, data: DataType = None) -> DataType:
        """
        Sends a message to the connected server and waits for a reply.

        Args:
            header (Header): The header to send.
            data (DataType, optional): The data to send.
        Returns:
            DataType: The reply data from the server.
        Raises:
            RemoteException: If the server replies with an error.
        """
        sequence_number = get_next_sequence_number()
        header.sequence_number = sequence_number
        response = asyncio.Future()

        def set_response(data: DataType) -> None:
            if response.done():
                return

            if isinstance(data, ErrorStr):
                self.logger.error(
                    "Received ERROR reply for sequence number %d",
                    sequence_number,
                )
                error_message = data if isinstance(data, str) else "Unknown error"
                response.set_exception(
                    RemoteException(f"Error from server: {error_message}")
                )
            else:
                response.set_result(data)

        self.once(f"reply-{sequence_number}", set_response)
        await self._send(header, data)
        return await response

    async def prop_get(self, prop: str) -> DataType:
        """
        Reads and returns the current value of a property from the remote host.
        Single-valued, associative-array and data-array types can be returned.

        Args:
            prop (str): The name of the property to get.
        Returns:
            DataType: The value of the property.
        Raises:
            RemoteException: If the property does not exist on the remote host, or another error occurs.
        """
        return await self._send_with_reply(Header(Command.CHAN_READ, name=prop))

    async def prop_set(self, prop: str, value: DataType) -> None:
        """
        Sets a property to a value on the remote host.
        Single-valued, associative-array and data-array types can be sent.

        Args:
            prop (str): The name of the property to set.
            value (DataType): The value to set the property to.
        Raises:
            RemoteException: If the property does not exist on the remote host, or another error occurs.
        """
        await self._send(Header(Command.CHAN_SEND, name=prop), data=value)

    async def prop_watch(self, prop: str) -> None:
        """
        Registers a property on the remote host for watching.
        When the property value changes, the remote host will send an event to the client.
        Consider:
            prop_watch("var/TEMP")
        If a variable named TEMP exists on the local client, then the value of the local client's instance
        will track changes to the value of the same variable on the remote host.

        The variable must exist on the server before the client requests it be watched.
        If the variable goes out of existence on the server, but is subsequently recreated as the same type of global variable,
        the watched status will be reinstated (as of spec release 5.05.05-1).

        If the variable doesn't exist on the client or goes out of existence, the client will continue to receive events,
        and if the variable is recreated on the client, its value will track the values sent with the events (as of spec release 5.05.05-1).

        Regular global variables, associative arrays and associative array elements can be watched.
        Data arrays cannot be watched.
        The built-in motor and scaler arrays A[] and S[] can be watched, but events will only be generated when
        the elements are explicitly assigned values on the server,
        not when the values change by way of built-in code, such as from calcA, getangles or getcounts.

        Args:
            prop (str): The name of the property to watch.
        """
        await self._send(Header(Command.REGISTER, name=prop))

    async def prop_unwatch(self, prop: str) -> None:
        """
        Unregisters a property on the remote host.
        The remote host will no longer send events to the client when the property value changes.

        Args:
            prop (str): The name of the property to unwatch.
        """
        await self._send(Header(Command.UNREGISTER, name=prop))

    async def abort(self) -> None:
        """
        Aborts the current command on the remote host.
        This has the same effect on the remote host as a ^C from the keyboard.
        Any pending commands in the server queue from the client will be removed.
        """
        await self._send(Header(Command.ABORT))

    @asynccontextmanager
    async def _abort_on_interrupt(self):
        """
        Context manager to automatically abort the current command on the remote host if an interrupt signal is received.
        """
        try:
            yield
        except (asyncio.CancelledError, KeyboardInterrupt, SystemExit):
            await self.abort()
            raise

    async def remote_cmd_no_return(self, cmd: str) -> None:
        """
        Puts the spec command on the execution queue of the remote host.
        Does not wait for the command to resolve or return a value.

        Args:
            cmd (str): The command string to send to the remote host. e.g. "1+1"
        """
        await self._send(Header(Command.CMD), data=cmd)

    async def remote_cmd(self, cmd: str) -> DataType:
        """
        Puts the spec command on the execution queue of the remote host.
        Waits for the command to resolve and returns the resulting value.

        Args:
            cmd (str): The command string to send to the remote host. e.g. "1+1"
        Returns:
            DataType: The result of the command execution from the remote host.
        """

        async with self._abort_on_interrupt():
            return await self._send_with_reply(
                Header(Command.CMD_WITH_RETURN), data=cmd
            )

    async def remote_func_no_return(self, func: str, *args) -> None:
        """
        Calls a function on the remote host with the provided arguments.
        Does not wait for the function to resolve or return a value.

        Args:
            func (str): The name of the function to call on the remote host.
            *args: The arguments to pass to the function. These will all be converted to strings before sending.
        """
        func_string = f"{func}(" + ", ".join(repr(arg) for arg in args) + ")"
        await self._send(Header(Command.FUNC), data=func_string)

    async def remote_func(self, func: str, *args) -> DataType:
        """
        Calls a function on the remote host with the provided arguments.
        Waits for the command to resolve and returns the resulting value.

        Args:
            func (str): The name of the function to call on the remote host.
            *args: The arguments to pass to the function. These will all be converted to strings before sending.
        Returns:
            DataType: The result of the function execution from the remote host.
        """
        func_string = f"{func}(" + ", ".join(repr(arg) for arg in args) + ")"
        async with self._abort_on_interrupt():
            return await self._send_with_reply(
                Header(Command.FUNC_WITH_RETURN), data=func_string
            )

    async def hello(self, *, timeout: float = 5.0):
        """
        Sends a HELLO command to the remote host.
        The remote host should respond with a HELLO_REPLY message.

        Waits for the reply up to the specified timeout.

        Args:
            timeout (float, optional): The maximum time to wait for a reply, in seconds.
        Returns:
            DataType: The message received from the server in response to the HELLO command.
        """
        return await asyncio.wait_for(
            self._send_with_reply(Header(Command.HELLO)),
            timeout=timeout,
        )

    @asynccontextmanager
    async def synchronized_motors(self, *, timeout: Optional[float] = None):
        """
        Context manager to enable synchronized motor operations for the client.

        While this context is active, motor movements will be held.
        Upon exiting the context, the movements will be initialized simultaneously.

        Example usage:

        .. code-block:: python

            async with client_connection.synchronized_motors():
                # Motor movement will be held in here.
                motor1.move(position)
                motor2.move(position)

                # Motors will not start moving yet.
                await asyncio.sleep(1)  # Simulate other operations
                # Motors will start moving simultaneously here.

            # Outside of the context, all motors have completed their movements.

        Args:
            timeout (float, optional): Maximum time to wait for all motors to complete, in seconds.
        Yields:
            None
        Raises:
            RuntimeError: If there are pending motor motions from a previous context.
        """
        assert (
            not self._synchronizing_motors
        ), "Concurrent synchronized_motors contexts are not allowed."

        move_done_pattern = re.compile(r"motor/(.+)/move_done")

        move_started: dict[str, bool] = {}
        waiting_for: dict[str, asyncio.Future] = {}

        def motor_move_done_check(name: str, value: DataType) -> None:
            if (match := move_done_pattern.match(name)) is None:
                return
            motor_name = match.group(1)

            if motor_name in waiting_for:
                if value:
                    # Look for if we have started a movement (move_done == True)
                    move_started[motor_name] = True
                elif move_started.get(motor_name, False):
                    # If we have started one for this motor, then we look for a subsequent move_done == False,
                    # which indicates the movement is complete.
                    #
                    # This latching system is so that we don't get confused since the motor typically
                    # starts with move_done == False, and if we just look for the first move_done == False,
                    # we might mistakenly think the motor has completed its movement before it even starts.
                    waiting_for[motor_name].set_result(True)

                self.logger.info(
                    "move_done=`%s` received for `%s` during synchronized motion.",
                    value,
                    motor_name,
                )

        motion_started = False
        try:
            if len(self._pending_motions) > 0:
                raise RuntimeError(
                    "There are pending motor motions from a previous synchronized_motors context."
                )
            self._synchronizing_motors = True

            # Give control back to user.
            yield

            self.on("property-change", motor_move_done_check)
            try:
                for mne in self._pending_motions.keys():
                    waiting_for[mne] = asyncio.Future()

                # Start the prestart message
                motion_started = True
                await self.prop_set("motor/../prestart_all", None)

                # Append the individual motor commands
                for mne, position in self._pending_motions.items():
                    self.logger.info(
                        "Starting synchronized move for `%s` to position %s.",
                        mne,
                        position,
                    )

                    await self.prop_set(f"motor/{mne}/start_one", position)

                # Start all the motors simultaneously
                await self.prop_set("motor/../start_all", None)

                # Wait for them to be done
                await asyncio.wait(waiting_for.values(), timeout=timeout)
            finally:
                self.remove_listener("property-change", motor_move_done_check)
        except Exception as e:
            self.logger.error("Error during synchronized motor operations: %s", str(e))
            if motion_started:
                await self.prop_set("motor/../abort_all", None)
            raise
        finally:
            self._synchronizing_motors = False
            self._pending_motions.clear()
